from collections.abc import Generator, Sequence
from typing import Any
from unittest.mock import MagicMock, patch

import pytest

from any_agent.config import AgentFramework, MCPSse, Tool
from any_agent.tools import _get_mcp_server


@pytest.fixture
def load_mcp_tools(
    tools: Sequence[Tool],
) -> Generator[MagicMock]:
    with patch(
        "any_agent.tools.mcp.frameworks.langchain.load_mcp_tools"
    ) as mock_load_tools:
        mock_load_tools.return_value = tools
        yield mock_load_tools


@pytest.mark.asyncio
@pytest.mark.usefixtures(
    "enter_context_with_transport_and_session", "_path_client_session"
)
async def test_langchain_mcp_sse_integration(
    mcp_sse_params_no_tools: MCPSse,
    session: Any,
    load_mcp_tools: Any,
) -> None:
    server = _get_mcp_server(mcp_sse_params_no_tools, AgentFramework.LANGCHAIN)

    await server._setup_tools()

    session.initialize.assert_called_once()

    load_mcp_tools.assert_called_once_with(session)



import pathlib
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from mktestdocs import check_md_file


# Note the use of `str`, makes for pretty output
@pytest.mark.parametrize("fpath", pathlib.Path("docs").glob("**/*.md"), ids=str)
def test_files_all(fpath: pathlib.Path) -> None:
    mock_agent = MagicMock()
    mock_create = MagicMock(return_value=mock_agent)

    mock_create_async = AsyncMock()
    with (
        patch("builtins.open", new_callable=MagicMock),
        patch(
            "any_agent.evaluation.evaluation_runner.save_evaluation_results",
            return_value=None,
        ),
        patch("any_agent.evaluation.evaluation_runner.EvaluationRunner.run"),
        patch("any_agent.AnyAgent.create", mock_create),
        patch("any_agent.AnyAgent.create_async", mock_create_async),
    ):
        check_md_file(fpath=fpath, memory=True)  # type: ignore[no-untyped-call]



import json
import re
from abc import ABC
from collections.abc import Sequence
from textwrap import dedent

from litellm import completion

from any_agent.evaluation.evaluation_case import CheckpointCriteria, GroundTruthAnswer
from any_agent.evaluation.evaluators.schemas import EvaluationResult


class LLMEvaluator(ABC):
    """Base class for evaluators that use LLM-as-judge."""

    def __init__(self, model: str):
        self.model = model

    def llm_evaluate_with_criterion(
        self,
        criteria: str,
        points: int,
        ground_truth_output: Sequence[CheckpointCriteria]
        | Sequence[GroundTruthAnswer]
        | None = None,
        hypothesis_final_output: str | None = None,
        evidence: str | None = None,
    ) -> EvaluationResult:
        """Evaluate a single criterion using LLM."""
        prompt = dedent(f"""
        Evaluate if the following criterion was met {"based on the provided evidence" if evidence else "in the agent's answer"}.

        Criterion: {criteria}
        """)

        if ground_truth_output:
            prompt += dedent(f"""
            Expected output: {json.dumps(ground_truth_output)}
            """)
        if hypothesis_final_output:
            prompt += dedent(f"""
            Agent's answer: {hypothesis_final_output}
            """)

        if evidence:
            prompt += dedent(f"""
            Trace evidence:
            {evidence}
            """)

        prompt += f"""

        Based on the {"evidence" if evidence else "comparison between the expected output and the actual final answer"},
        was this criterion satisfied? Answer with:
        1. "passed": true or false
        2. "reason": Brief explanation for your decision
        """
        prompt += """
        Output valid JSON with these three fields only, in the format:
        ```json
        {
            "passed": true,
            "reason": "I have them"
        }
        ```
        """

        response = completion(
            model=self.model,
            messages=[{"role": "user", "content": prompt}],
        )
        content = response.choices[0].message.content

        try:
            # Extract JSON from the response - looks for patterns like ```json {...} ``` or just {...}
            json_match = re.search(
                r"```(?:json)?\s*(\{.*?\})\s*```|(\{.*?\})",
                content,
                re.DOTALL,
            )

            if json_match:
                # Use the first matching group that captured content
                json_str = next(group for group in json_match.groups() if group)
                evaluation = json.loads(json_str)
            else:
                # Fallback: try parsing the whole content as JSON
                evaluation = json.loads(content)

            evaluation["criteria"] = criteria
        except (json.JSONDecodeError, AttributeError, StopIteration) as e:
            evaluation = {
                "passed": False,
                "reason": f"Failed to evaluate due to parsing: {e!s} \n Response: {content}",
                "criteria": criteria,
            }
        evaluation["points"] = points
        return EvaluationResult.model_validate(evaluation)



from pathlib import Path
from unittest.mock import MagicMock, patch

from any_agent.config import AgentFramework
from any_agent.evaluation.evaluation_case import EvaluationCase
from any_agent.evaluation.evaluation_runner import EvaluationRunner
from any_agent.evaluation.evaluators.schemas import EvaluationResult
from any_agent.tracing.trace import AgentTrace


def test_runner_add_case(evaluation_case: EvaluationCase, tmp_path: Path) -> None:
    output_path = str(tmp_path / "test_output")
    runner = EvaluationRunner(output_path)

    # Add the test case
    runner.add_evaluation_case(evaluation_case)

    # Try to add the same case again
    runner.add_evaluation_case(evaluation_case)

    # Check that the case is still in the list
    assert len(runner._evaluation_cases) == 1, (
        "Evaluation case should only be added once."
    )

    # change the case and add again

    new_case = evaluation_case.model_copy(deep=True)
    new_case.ground_truth[0]["value"] = 2.0
    runner.add_evaluation_case(new_case)
    assert len(runner._evaluation_cases) == 2, (
        "Evaluation case should be added again with different values."
    )


def test_evaluation_runner_add_trace(agent_trace: AgentTrace, tmp_path: Path) -> None:
    output_path = str(tmp_path / "test_output")
    runner = EvaluationRunner(output_path)

    runner.add_trace(agent_trace, AgentFramework.OPENAI)
    assert len(runner._traces) == 1, "Trace should be added to the runner."

    runner.add_trace(agent_trace, AgentFramework.OPENAI)
    assert len(runner._traces) == 1, "Trace should not be added again."

    second_trace = agent_trace.model_copy(deep=True)
    second_trace.spans[0].name = "Different Span"
    runner.add_trace(second_trace, AgentFramework.OPENAI)
    assert len(runner._traces) == 2, "Second trace should be added to the runner."


def test_evaluation_runner_runs_all_cases(
    evaluation_case: EvaluationCase, agent_trace: AgentTrace, tmp_path: Path
) -> None:
    """This unit test is designed to ensure that the EvaluationRunner properly iterates over all the traces and eval cases"""

    ### Set up the Runner and add the traces and eval cases
    output_path = str(tmp_path / "test_output")
    runner = EvaluationRunner(output_path)

    second_case = evaluation_case.model_copy(deep=True)
    second_case.ground_truth[0]["value"] = 2.0

    runner.add_evaluation_case(evaluation_case)
    runner.add_evaluation_case(second_case)

    # Add multiple traces
    second_trace = agent_trace.model_copy(deep=True)
    second_trace.spans[0].name = "Different Span"

    runner.add_trace(agent_trace, AgentFramework.OPENAI)
    runner.add_trace(second_trace, AgentFramework.OPENAI)

    #### Set up the mocks for the evaluators so that we don't actually call LLMs.
    mock_checkpoint_evaluator1 = MagicMock()
    mock_checkpoint_evaluator2 = MagicMock()
    mock_hypothesis_evaluator1 = MagicMock()
    mock_hypothesis_evaluator2 = MagicMock()
    mock_qa_evaluator1 = MagicMock()
    mock_qa_evaluator2 = MagicMock()

    ### Every evaluator will return the same result
    eval_result = [
        EvaluationResult(
            criteria="test criteria", passed=True, reason="test passed", points=1
        )
    ]
    for evaluator in [
        mock_checkpoint_evaluator1,
        mock_checkpoint_evaluator2,
        mock_hypothesis_evaluator1,
        mock_hypothesis_evaluator2,
        mock_qa_evaluator1,
        mock_qa_evaluator2,
    ]:
        evaluator.evaluate.return_value = eval_result

    with (
        patch(
            "any_agent.evaluation.evaluation_runner.CheckpointEvaluator"
        ) as mock_checkpoint_eval_cls,
        patch(
            "any_agent.evaluation.evaluation_runner.HypothesisEvaluator"
        ) as mock_hypothesis_eval_cls,
        patch(
            "any_agent.evaluation.evaluation_runner.QuestionAnsweringSquadEvaluator"
        ) as mock_qa_eval_cls,
        patch(
            "any_agent.evaluation.evaluation_runner.save_evaluation_results"
        ) as mock_save_results,
        patch(
            "any_agent.evaluation.evaluation_runner.TracingProcessor.create"
        ) as mock_processor_create,
    ):
        # Evaluators are created for each evaluation case (but not for each trace)
        mock_checkpoint_eval_cls.side_effect = [
            mock_checkpoint_evaluator1,
            mock_checkpoint_evaluator2,
        ]
        mock_hypothesis_eval_cls.side_effect = [
            mock_hypothesis_evaluator1,
            mock_hypothesis_evaluator2,
        ]
        mock_qa_eval_cls.side_effect = [mock_qa_evaluator1, mock_qa_evaluator2]

        # Mock processor is used to extract the hypothesis answer
        mock_processor = MagicMock()
        mock_processor._extract_hypothesis_answer.return_value = (
            "Mock hypothesis answer"
        )
        mock_processor_create.return_value = mock_processor

        # Run evaluation
        runner.run()

        # Verify that each evaluator class was instantiated twice (once per evaluation case)
        assert mock_checkpoint_eval_cls.call_count == 2
        assert mock_hypothesis_eval_cls.call_count == 2
        assert mock_qa_eval_cls.call_count == 2

        # Verify that each evaluator instance's evaluate method was called twice (once per trace)
        assert mock_checkpoint_evaluator1.evaluate.call_count == 2
        assert mock_checkpoint_evaluator2.evaluate.call_count == 2
        assert mock_hypothesis_evaluator1.evaluate.call_count == 2
        assert mock_hypothesis_evaluator2.evaluate.call_count == 2
        assert mock_qa_evaluator1.evaluate.call_count == 2
        assert mock_qa_evaluator2.evaluate.call_count == 2

        # Verify that save_evaluation_results was called once per trace per case (2 cases * 2 traces = 4 times)
        assert mock_save_results.call_count == 4



import os
from pathlib import Path

import pandas as pd

from any_agent.evaluation.evaluation_case import EvaluationCase
from any_agent.evaluation.results_saver import save_evaluation_results
from any_agent.tracing.trace import AgentTrace


def test_save_evaluation_results_creates_file(
    evaluation_case: EvaluationCase, agent_trace: AgentTrace, tmp_path: Path
) -> None:
    output_path = str(tmp_path / "test_output.json")

    save_evaluation_results(
        evaluation_case=evaluation_case,
        output_path=output_path,
        output_message="Test evaluation completed.",
        trace=agent_trace,
        hypothesis_answer="This is a test answer.",
        passed_checks=1,
        failed_checks=0,
        score=100.0,
    )

    assert os.path.exists(output_path), "Output file was not created."

    # Clean up after test
    os.remove(output_path)


def test_save_evaluation_results_writes_correct_data(
    evaluation_case: EvaluationCase, agent_trace: AgentTrace, tmp_path: Path
) -> None:
    output_path = str(tmp_path / "test_output.json")
    if os.path.exists(output_path):
        os.remove(output_path)  # Clean up before test

    save_evaluation_results(
        evaluation_case=evaluation_case,
        output_path=output_path,
        output_message="Test evaluation completed.",
        trace=agent_trace,
        hypothesis_answer="This is a test answer.",
        passed_checks=1,
        failed_checks=0,
        score=100.0,
    )

    df = pd.read_json(output_path, orient="records", lines=True)

    assert len(df) == 1, "DataFrame should contain one record."
    assert df.iloc[0]["hypothesis_answer"] == "This is a test answer."
    assert df.iloc[0]["passed_checks"] == 1
    assert df.iloc[0]["failed_checks"] == 0
    assert df.iloc[0]["score"] == 100.0

    # Clean up after test
    os.remove(output_path)



from typing import TYPE_CHECKING, Any

from any_agent.config import AgentConfig, AgentFramework, TracingConfig
from any_agent.frameworks.any_agent import AnyAgent
from any_agent.tools import search_web, visit_webpage

try:
    from smolagents import LiteLLMModel, ToolCallingAgent

    smolagents_available = True
except ImportError:
    smolagents_available = False

if TYPE_CHECKING:
    from smolagents import MultiStepAgent

    from any_agent.tracing.trace import AgentTrace


DEFAULT_AGENT_TYPE = ToolCallingAgent
DEFAULT_MODEL_TYPE = LiteLLMModel


class SmolagentsAgent(AnyAgent):
    """Smolagents agent implementation that handles both loading and running."""

    def __init__(
        self,
        config: AgentConfig,
        managed_agents: list[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ):
        super().__init__(config, managed_agents, tracing)
        self._agent: MultiStepAgent | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.SMOLAGENTS

    def _get_model(self, agent_config: AgentConfig) -> Any:
        """Get the model configuration for a smolagents agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE
        model_args = agent_config.model_args or {}
        kwargs = {
            "model_id": agent_config.model_id,
            "api_key": agent_config.api_key,
            "api_base": agent_config.api_base,
            **model_args,
        }
        return model_type(**kwargs)

    async def _load_agent(self) -> None:
        """Load the Smolagents agent with the given configuration."""
        if not smolagents_available:
            msg = "You need to `pip install 'any-agent[smolagents]'` to use this agent"
            raise ImportError(msg)
        if not self.managed_agents and not self.config.tools:
            self.config.tools = [
                search_web,
                visit_webpage,
            ]

        tools, _ = await self._load_tools(self.config.tools)

        managed_agents_instanced = []
        if self.managed_agents:
            for managed_agent in self.managed_agents:
                agent_type = managed_agent.agent_type or DEFAULT_AGENT_TYPE
                managed_tools, _ = await self._load_tools(managed_agent.tools)
                managed_agent_instance = agent_type(
                    name=managed_agent.name,
                    model=self._get_model(managed_agent),
                    tools=managed_tools,
                    verbosity_level=-1,  # OFF
                    description=managed_agent.description
                    or f"Use the agent: {managed_agent.name}",
                )
                if managed_agent.instructions:
                    managed_agent_instance.prompt_templates["system_prompt"] = (
                        managed_agent.instructions
                    )
                managed_agents_instanced.append(managed_agent_instance)

        main_agent_type = self.config.agent_type or DEFAULT_AGENT_TYPE

        self._agent = main_agent_type(
            name=self.config.name,
            model=self._get_model(self.config),
            tools=tools,
            verbosity_level=-1,  # OFF
            managed_agents=managed_agents_instanced,
            **self.config.agent_args or {},
        )

        assert self._agent

        if self.config.instructions:
            self._agent.prompt_templates["system_prompt"] = self.config.instructions

    async def run_async(self, prompt: str, **kwargs: Any) -> "AgentTrace":
        """Run the Smolagents agent with the given prompt."""
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        self._setup_tracing()
        result = self._agent.run(prompt, **kwargs)
        self._exporter.trace.final_output = result
        return self._exporter.trace



import time
from collections.abc import AsyncGenerator, Generator
from textwrap import dedent
from unittest.mock import AsyncMock, patch

import pytest
import rich.console
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.trace import SpanContext, SpanKind, TraceFlags, TraceState
from opentelemetry.trace.status import Status, StatusCode

from any_agent.config import AgentFramework


@pytest.fixture(autouse=True)
def disable_rich_console(
    monkeypatch: pytest.MonkeyPatch,
    pytestconfig: pytest.Config,
) -> None:
    original_init = rich.console.Console.__init__

    def quiet_init(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        if pytestconfig.option.capture != "no":
            kwargs["quiet"] = True
        original_init(self, *args, **kwargs)

    monkeypatch.setattr(rich.console.Console, "__init__", quiet_init)


@pytest.fixture
def llm_span() -> ReadableSpan:
    # Convert hex trace and span IDs to integers
    trace_id = int("69ea1b41a9bc5724381993def669c803", 16)
    span_id = int("3817abcfb97cc40c", 16)
    parent_span_id = int("68ff5f13e03ac3fd", 16)

    context = SpanContext(
        trace_id=trace_id,
        span_id=span_id,
        is_remote=False,
        trace_flags=TraceFlags(TraceFlags.SAMPLED),
        trace_state=TraceState(),
    )

    parent = SpanContext(
        trace_id=trace_id,
        span_id=parent_span_id,
        is_remote=False,
        trace_flags=TraceFlags(TraceFlags.SAMPLED),
        trace_state=TraceState(),
    )

    resource = Resource.create(
        {
            "telemetry.sdk.language": "python",
            "telemetry.sdk.name": "opentelemetry",
            "telemetry.sdk.version": "1.32.0",
            "service.name": "unknown_service",
        }
    )
    # Create a ReadableSpan
    return ReadableSpan(
        name="ChatLiteLLM",
        context=context,
        kind=SpanKind.INTERNAL,
        parent=parent,
        start_time=int(time.time()),
        end_time=int(time.time() + 1),
        status=Status(StatusCode.OK),
        attributes={
            "input.value": '{"messages": [[{"lc": 1, "type": "constructor", "id": ["langchain", "schema", "messages", "SystemMessage"], "kwargs": {"content": "Use the available tools to find the answer", "type": "system"}}, {"lc": 1, "type": "constructor", "id": ["langchain", "schema", "messages", "HumanMessage"], "kwargs": {"content": "Which agent framework is the best?", "type": "human", "id": "2aaf3de6-edf7-4cfa-9483-da348a6749da"}}]]}',
            "input.mime_type": "application/json",
            "output.value": '{"generations": [[{"text": "", "generation_info": {"finish_reason": "tool_calls"}, "type": "ChatGeneration", "message": {"lc": 1, "type": "constructor", "id": ["langchain", "schema", "messages", "AIMessage"], "kwargs": {"content": "", "additional_kwargs": {"tool_calls": [{"lc": 1, "type": "not_implemented", "id": ["litellm", "types", "utils", "ChatCompletionMessageToolCall"], "repr": "ChatCompletionMessageToolCall(function=Function(arguments=\'{\\"query\\":\\"best agent frameworks 2023\\"}\', name=\'search_web\'), id=\'call_hYPpXiyNNLiRRytLlJdNFpGN\', type=\'function\')"}]}, "response_metadata": {"token_usage": {"lc": 1, "type": "not_implemented", "id": ["litellm", "types", "utils", "Usage"], "repr": "Usage(completion_tokens=20, prompt_tokens=145, total_tokens=165, completion_tokens_details=CompletionTokensDetailsWrapper(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0, text_tokens=None), prompt_tokens_details=PromptTokensDetailsWrapper(audio_tokens=0, cached_tokens=0, text_tokens=None, image_tokens=None))"}, "model": "gpt-4o-mini", "finish_reason": "tool_calls", "model_name": "gpt-4o-mini"}, "type": "ai", "id": "run-984943f2-6546-47fc-9b1d-81714109e374-0", "tool_calls": [{"name": "search_web", "args": {"query": "best agent frameworks 2023"}, "id": "call_hYPpXiyNNLiRRytLlJdNFpGN", "type": "tool_call"}], "usage_metadata": {"input_tokens": 145, "output_tokens": 20, "total_tokens": 165}, "invalid_tool_calls": []}}}]], "llm_output": {"token_usage": {"completion_tokens": 20, "prompt_tokens": 145, "total_tokens": 165, "completion_tokens_details": {"accepted_prediction_tokens": 0, "audio_tokens": 0, "reasoning_tokens": 0, "rejected_prediction_tokens": 0}, "prompt_tokens_details": {"audio_tokens": 0, "cached_tokens": 0}}, "model": "gpt-4o-mini"}, "run": null, "type": "LLMResult"}',
            "output.mime_type": "application/json",
            "llm.input_messages.0.message.role": "system",
            "llm.input_messages.0.message.content": "Use the available tools to find the answer",
            "llm.input_messages.1.message.role": "user",
            "llm.input_messages.1.message.content": "Which agent framework is the best?",
            "llm.output_messages.0.message.role": "assistant",
            "llm.invocation_parameters": '{"model": "gpt-4o-mini", "temperature": null, "top_p": null, "top_k": null, "n": null, "_type": "litellm-chat", "stop": null, "tools": [{"type": "function", "function": {"name": "search_web", "description": "Performs a duckduckgo web search based on your query (think a Google search) then returns the top search results.\\n\\n    Args:\\n        query (str): The search query to perform.\\n\\n    Returns:\\n        The top search results.", "parameters": {"properties": {"query": {"type": "string"}}, "required": ["query"], "type": "object"}}}, {"type": "function", "function": {"name": "visit_webpage", "description": "Visits a webpage at the given url and reads its content as a markdown string. Use this to browse webpages.\\n\\n    Args:\\n        url: The url of the webpage to visit.", "parameters": {"properties": {"url": {"type": "string"}}, "required": ["url"], "type": "object"}}}], "tool_choice": null}',
            "llm.model_name": "gpt-4o-mini",
            "llm.token_count.prompt": 145,
            "llm.token_count.completion": 20,
            "llm.token_count.total": 165,
            "metadata": '{"langgraph_step": 1, "langgraph_node": "agent", "langgraph_triggers": ["branch:to:agent", "start:agent", "tools"], "langgraph_path": ["__pregel_pull", "agent"], "langgraph_checkpoint_ns": "agent:b43d5300-228d-6e29-4b2f-7994f7924340", "checkpoint_ns": "agent:b43d5300-228d-6e29-4b2f-7994f7924340", "ls_provider": "litellm", "ls_model_type": "chat", "ls_model_name": "gpt-4o-mini"}',
            "openinference.span.kind": "LLM",
        },
        events=[],
        links=[],
        resource=resource,
    )


@pytest.fixture(params=list(AgentFramework), ids=lambda x: x.name)
def agent_framework(request: pytest.FixtureRequest) -> AgentFramework:
    return request.param  # type: ignore[no-any-return]


@pytest.fixture
def _patch_stdio_client() -> Generator[
    tuple[AsyncMock, tuple[AsyncMock, AsyncMock]], None
]:
    mock_cm = AsyncMock()
    mock_transport = (AsyncMock(), AsyncMock())
    mock_cm.__aenter__.return_value = mock_transport

    with patch("mcp.client.stdio.stdio_client", return_value=mock_cm) as patched:
        yield patched, mock_transport


SSE_MCP_SERVER_SCRIPT = dedent(
    '''
        from mcp.server.fastmcp import FastMCP

        mcp = FastMCP("Echo Server", host="127.0.0.1", port=8000)

        @mcp.tool()
        def write_file(text: str) -> str:
            """Say hi back with the input text"""
            return f"Hi: {text}"

        @mcp.tool()
        def read_file(text: str) -> str:
            """Say bye back the input text"""
            return f"Bye: {text}"

        @mcp.tool()
        def other_tool(text: str) -> str:
            """Say boo back the input text"""
            return f"Boo: {text}"

        mcp.run("sse")
        '''
)


@pytest.fixture(
    scope="session"
)  # This means it only gets created once per test session
async def echo_sse_server() -> AsyncGenerator[dict[str, str]]:
    """This fixture runs a FastMCP server in a subprocess.
    I thought about trying to mock all the individual mcp client calls,
    but I went with this because this way we don't need to actually mock anything.
    This is similar to what MCPAdapt does in their testing https://github.com/grll/mcpadapt/blob/main/tests/test_core.py
    """
    import asyncio

    process = await asyncio.create_subprocess_exec(
        "python",
        "-c",
        SSE_MCP_SERVER_SCRIPT,
    )
    await asyncio.sleep(1)

    try:
        yield {"url": "http://127.0.0.1:8000/sse"}
    finally:
        # Clean up the process when test is done
        process.kill()
        await process.wait()



import json

from fire import Fire

from any_agent.config import AgentFramework
from any_agent.evaluation import EvaluationRunner
from any_agent.evaluation.evaluation_case import EvaluationCase
from any_agent.logging import logger
from any_agent.tracing.trace import AgentSpan, AgentTrace


def do_eval(
    evaluation_case_paths: list[str],
    trace_paths: list[str],
    agent_framework: AgentFramework,
    output_path: str = "output/results.json",
) -> None:
    logger.info("Starting evaluation...")
    runner = EvaluationRunner(output_path=output_path)

    for evaluation_case_path in evaluation_case_paths:
        logger.info("Loading test case from %s", evaluation_case_path)
        evaluation_case = EvaluationCase.from_yaml(evaluation_case_path)
        runner.add_evaluation_case(evaluation_case)

    for trace_path in trace_paths:
        logger.info("Loading tracing from %s", trace_path)
        with open(trace_path, encoding="utf-8") as f:
            spans = json.loads(f.read())
        spans = [AgentSpan.model_validate_json(span) for span in spans]
        trace = AgentTrace(spans=spans)
        # dump the trace to a file
        with open("tmp.tmp", "w", encoding="utf-8") as f:
            f.write(trace.model_dump_json(indent=2))
        runner.add_trace(trace, agent_framework)

    logger.info("Running evaluation...")
    runner.run()
    logger.info("Evaluation completed.")


def main() -> None:
    Fire(do_eval)  # type: ignore[no-untyped-call]


if __name__ == "__main__":
    main()



from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.tools import (
    search_web,
    visit_webpage,
)


def test_load_langchain_agent_default() -> None:
    model_mock = MagicMock()
    create_mock = MagicMock()
    agent_mock = MagicMock()
    create_mock.return_value = agent_mock
    tool_mock = MagicMock()

    with (
        patch("any_agent.frameworks.langchain.DEFAULT_AGENT_TYPE", create_mock),
        patch("any_agent.frameworks.langchain.DEFAULT_MODEL_TYPE", model_mock),
        patch("langchain_core.tools.tool", tool_mock),
    ):
        AnyAgent.create(AgentFramework.LANGCHAIN, AgentConfig(model_id="gpt-4o"))

        model_mock.assert_called_once_with(model="gpt-4o", api_base=None, api_key=None)
        create_mock.assert_called_once_with(
            name="any_agent",
            model=model_mock.return_value,
            tools=[tool_mock(search_web), tool_mock(visit_webpage)],
            prompt=None,
        )


def test_load_langchain_agent_missing() -> None:
    with patch("any_agent.frameworks.langchain.langchain_available", False):
        with pytest.raises(ImportError):
            AnyAgent.create(AgentFramework.LANGCHAIN, AgentConfig(model_id="gpt-4o"))


def test_load_langchain_multiagent() -> None:
    model_mock = MagicMock()
    create_mock = MagicMock()
    handoff_mock = MagicMock()

    def _create_effect(**kwargs: str) -> MagicMock:
        agent_mock = MagicMock()
        agent_mock.name = kwargs["name"]
        return agent_mock

    create_mock.side_effect = _create_effect
    tool_mock = MagicMock()

    with (
        patch("any_agent.frameworks.langchain.DEFAULT_AGENT_TYPE", create_mock),
        patch("any_agent.frameworks.langchain.create_handoff_tool", handoff_mock),
        patch("any_agent.frameworks.langchain.DEFAULT_MODEL_TYPE", model_mock),
        patch("langchain_core.tools.tool", tool_mock),
    ):
        main_agent = AgentConfig(
            model_id="o3-mini",
        )

        managed_agents = [
            AgentConfig(
                model_id="gpt-4o",
                tools=[
                    search_web,
                    visit_webpage,
                ],
            ),
        ]

        AnyAgent.create(
            AgentFramework.LANGCHAIN,
            main_agent,
            managed_agents=managed_agents,
        )

        create_mock.assert_any_call(
            name="managed_agent_0",
            model=model_mock.return_value,
            tools=[
                tool_mock(search_web),
                tool_mock(visit_webpage),
                handoff_mock.return_value,
            ],
            prompt=None,
        )

        create_mock.assert_any_call(
            name="any_agent",
            model=model_mock.return_value,
            tools=[handoff_mock.return_value],
            prompt=None,
        )


def test_run_langchain_agent_custom_args() -> None:
    create_mock = MagicMock()
    agent_mock = AsyncMock()
    agent_mock.ainvoke.return_value = MagicMock()
    create_mock.return_value = agent_mock

    with (
        patch("any_agent.frameworks.langchain.DEFAULT_AGENT_TYPE", create_mock),
        patch("any_agent.frameworks.langchain.DEFAULT_MODEL_TYPE"),
        patch("langchain_core.tools.tool"),
    ):
        agent = AnyAgent.create(
            AgentFramework.LANGCHAIN, AgentConfig(model_id="gpt-4o")
        )
        agent.run("foo", debug=True)
        agent_mock.ainvoke.assert_called_once_with(
            {"messages": [("user", "foo")]}, debug=True
        )



from pydantic import TypeAdapter

from any_agent.config import AgentFramework, MCPParams

from .agno import AgnoMCPServer
from .google import GoogleMCPServer
from .langchain import LangchainMCPServer
from .llama_index import LlamaIndexMCPServer
from .openai import OpenAIMCPServer
from .smolagents import SmolagentsMCPServer
from .tinyagent import TinyAgentMCPServer

MCPServer = (
    AgnoMCPServer
    | GoogleMCPServer
    | LangchainMCPServer
    | LlamaIndexMCPServer
    | OpenAIMCPServer
    | SmolagentsMCPServer
    | TinyAgentMCPServer
)


def _get_mcp_server(mcp_tool: MCPParams, agent_framework: AgentFramework) -> MCPServer:
    return TypeAdapter[MCPServer](MCPServer).validate_python(
        {"mcp_tool": mcp_tool, "framework": agent_framework}
    )


__all__ = [
    "AgnoMCPServer",
    "GoogleMCPServer",
    "LangchainMCPServer",
    "LlamaIndexMCPServer",
    "MCPServer",
    "OpenAIMCPServer",
    "SmolagentsMCPServer",
    "TinyAgentMCPServer",
    "_get_mcp_server",
]



from .mcp import (
    AgnoMCPServer,
    GoogleMCPServer,
    LangchainMCPServer,
    LlamaIndexMCPServer,
    MCPServer,
    OpenAIMCPServer,
    SmolagentsMCPServer,
    TinyAgentMCPServer,
    _get_mcp_server,
    _MCPConnection,
    _MCPServerBase,
)
from .user_interaction import (
    ask_user_verification,
    send_console_message,
    show_final_output,
    show_plan,
)
from .web_browsing import search_web, visit_webpage

__all__ = [
    "AgnoMCPServer",
    "GoogleMCPServer",
    "LangchainMCPServer",
    "LlamaIndexMCPServer",
    "MCPServer",
    "OpenAIMCPServer",
    "SmolagentsMCPServer",
    "TinyAgentMCPServer",
    "_MCPConnection",
    "_MCPServerBase",
    "_get_mcp_server",
    "ask_user_verification",
    "search_web",
    "send_console_message",
    "show_final_output",
    "show_plan",
    "visit_webpage",
]



from collections.abc import Sequence

import evaluate.loading
from typing_extensions import TypedDict

from any_agent.evaluation.evaluation_case import GroundTruthAnswer
from any_agent.evaluation.evaluators.schemas import EvaluationResult


class AnswerDetails(TypedDict):
    answer_start: Sequence[int]
    text: Sequence[str]


class GroundTruthAnswers(TypedDict):
    id: str
    answers: AnswerDetails


class QuestionAnsweringSquadEvaluator:
    """Directly compares answers without using LLM-as-judge."""

    def __init__(self) -> None:
        self.metric = evaluate.loading.load("squad")

    def evaluate(
        self,
        hypothesis_answer: str,
        ground_truth_answer: Sequence[GroundTruthAnswer],
    ) -> list[EvaluationResult]:
        """Directly compare answers using simple matching."""
        # format the answers so that they're dicts with 'id' and 'prediction' keys for hypo
        # and the ref has id and answers keys
        hypothesis_answers = [{"id": "1", "prediction_text": hypothesis_answer}]
        ground_truth_answers: list[GroundTruthAnswers] = [
            {
                "id": "1",
                "answers": {
                    "answer_start": [0],
                    "text": [str(ground_truth_answer[0]["value"])],
                },
            },
        ]
        # Use the SQuAD metric to compare answers
        result = self.metric.compute(
            predictions=hypothesis_answers,
            references=ground_truth_answers,
        )

        assert result, "The result of the evaluation is empty"

        match = EvaluationResult(
            passed=int(result["exact_match"]) == 1,
            reason=f"Partial Match (F1) score is {round(result['f1'], 2)}",
            criteria="Is the answer a direct match?",
            points=1,
        )
        return [match]



from collections.abc import Sequence

from any_agent.evaluation.evaluation_case import CheckpointCriteria, GroundTruthAnswer
from any_agent.evaluation.evaluators.LLMEvaluator import LLMEvaluator
from any_agent.evaluation.evaluators.schemas import EvaluationResult


class HypothesisEvaluator(LLMEvaluator):
    """Evaluates the final answer against ground truth."""

    def evaluate(
        self,
        hypothesis_final_output: str,
        ground_truth_answer_dict: Sequence[GroundTruthAnswer],
        ground_truth_checkpoints: Sequence[CheckpointCriteria],
    ) -> list[EvaluationResult]:
        """Verify if the final answer meets all specified criteria."""
        results = []

        for criterion in ground_truth_checkpoints:
            evaluation = self.llm_evaluate_with_criterion(
                criteria=criterion.criteria,
                points=criterion.points,
                ground_truth_output=ground_truth_answer_dict,
                hypothesis_final_output=hypothesis_final_output,
            )

            results.append(evaluation)

        return results



# pylint: disable=unused-argument, unused-variable
from collections.abc import Sequence
from typing import Any

import pytest

from any_agent.config import AgentFramework, MCPParams, MCPSse, Tool
from any_agent.tools import _get_mcp_server, _MCPConnection


@pytest.mark.asyncio
async def test_sse_tool_filtering(
    agent_framework: AgentFramework,
    sse_params_echo_server: MCPSse,
    tools: Sequence[str],
) -> None:
    server = _get_mcp_server(sse_params_echo_server, agent_framework)
    await server._setup_tools()
    if agent_framework is AgentFramework.AGNO:
        # Check that only the specified tools are included
        assert set(server.tools[0].functions.keys()) == set(tools)  # type: ignore[union-attr]
    else:
        assert len(server.tools) == len(tools)  # ignore[arg-type]


@pytest.mark.asyncio
async def test_mcp_tools_loaded(
    agent_framework: AgentFramework,
    mcp_params: MCPParams,
    mcp_connection: _MCPConnection[Any],
    tools: Sequence[Tool],
) -> None:
    mcp_server = _get_mcp_server(mcp_params, agent_framework)

    await mcp_server._setup_tools(mcp_connection=mcp_connection)

    assert mcp_server.tools == list(tools)



from collections.abc import Sequence
from typing import TYPE_CHECKING

from any_agent.evaluation.evaluation_case import CheckpointCriteria
from any_agent.evaluation.evaluators.LLMEvaluator import LLMEvaluator
from any_agent.evaluation.evaluators.schemas import EvaluationResult
from any_agent.logging import logger
from any_agent.tracing.processors.base import TracingProcessor

if TYPE_CHECKING:
    from any_agent.tracing.trace import AgentTrace


class CheckpointEvaluator(LLMEvaluator):
    """Evaluates checkpoints against trace."""

    def evaluate(
        self,
        trace: "AgentTrace",
        checkpoints: Sequence[CheckpointCriteria],
        processor: TracingProcessor,
    ) -> list[EvaluationResult]:
        """Verify each checkpoint against the trace data using LLM.

        Args:
            trace: The trace data to evaluate
            checkpoints: List of checkpoint criteria to verify
            processor: Trace processor to extract evidence

        Returns:
            List of evaluation results

        """
        evidence = processor.extract_evidence(trace)
        evidence = evidence.replace("<", "\\<").replace(">", "\\>")
        logger.info(f"""<yellow>Evidence\n{evidence}</yellow>\n""")
        results = []

        for checkpoint in checkpoints:
            evaluation = self.llm_evaluate_with_criterion(
                criteria=checkpoint.criteria,
                points=checkpoint.points,
                evidence=evidence,
            )
            results.append(evaluation)

        return results



from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.tools import (
    search_web,
    show_final_output,
    visit_webpage,
)


def test_load_google_default() -> None:
    from google.adk.tools import FunctionTool

    mock_agent = MagicMock()
    mock_model = MagicMock()
    mock_function_tool = MagicMock()

    class MockedFunctionTool(FunctionTool):
        def __new__(cls, *args: Any, **kwargs: Any) -> MagicMock:
            return mock_function_tool

    with (
        patch("any_agent.frameworks.google.LlmAgent", mock_agent),
        patch("any_agent.frameworks.google.DEFAULT_MODEL_TYPE", mock_model),
        patch("google.adk.tools.FunctionTool", MockedFunctionTool),
    ):
        AnyAgent.create(AgentFramework.GOOGLE, AgentConfig(model_id="gpt-4o"))
        mock_agent.assert_called_once_with(
            name="any_agent",
            instruction="",
            model=mock_model(model="gpt-4o"),
            tools=[MockedFunctionTool(search_web), MockedFunctionTool(visit_webpage)],
            sub_agents=[],
            output_key="response",
        )


def test_load_google_multiagent() -> None:
    from google.adk.tools import FunctionTool

    mock_agent = MagicMock()
    mock_model = MagicMock()
    mock_agent_tool = MagicMock()
    mock_function_tool = MagicMock()

    class MockedFunctionTool(FunctionTool):
        def __new__(cls, *args: Any, **kwargs: Any) -> "MockedFunctionTool":
            return mock_function_tool

    with (
        patch("any_agent.frameworks.google.LlmAgent", mock_agent),
        patch("any_agent.frameworks.google.DEFAULT_MODEL_TYPE", mock_model),
        patch("any_agent.frameworks.google.AgentTool", mock_agent_tool),
        patch("google.adk.tools.FunctionTool", MockedFunctionTool),
    ):
        AnyAgent.create(
            AgentFramework.GOOGLE,
            AgentConfig(model_id="gpt-4o"),
            managed_agents=[
                AgentConfig(
                    model_id="gpt-4o-mini",
                    name="search-web-agent",
                    tools=[
                        search_web,
                        visit_webpage,
                    ],
                ),
                AgentConfig(
                    model_id="gpt-4o-mini",
                    name="communication-agent",
                    tools=[show_final_output],
                    agent_args={"handoff": True},
                ),
            ],
        )

        mock_agent.assert_any_call(
            model=mock_model(model="gpt-4o-mini"),
            instruction="",
            name="search-web-agent",
            tools=[MockedFunctionTool(search_web), MockedFunctionTool(visit_webpage)],
        )
        mock_agent.assert_any_call(
            model=mock_model(model="gpt-4o-mini"),
            instruction="",
            name="communication-agent",
            tools=[MockedFunctionTool(show_final_output)],
        )
        mock_agent.assert_any_call(
            name="any_agent",
            instruction="",
            model=mock_model(model="gpt-4o"),
            tools=[mock_agent_tool.return_value],
            sub_agents=[mock_agent.return_value],
            output_key="response",
        )


def test_load_google_agent_missing() -> None:
    with patch("any_agent.frameworks.google.adk_available", False):
        with pytest.raises(ImportError):
            AnyAgent.create(AgentFramework.GOOGLE, AgentConfig(model_id="gpt-4o"))


def test_run_google_custom_args() -> None:
    from google.adk.agents.run_config import RunConfig
    from google.genai import types

    mock_agent = MagicMock()
    mock_runner = MagicMock()
    mock_runner.session_service.get_session.return_value = AsyncMock()

    run_config = RunConfig(max_llm_calls=10)
    with (
        patch("any_agent.frameworks.google.LlmAgent", mock_agent),
        patch("any_agent.frameworks.google.InMemoryRunner", mock_runner),
        patch("any_agent.frameworks.google.DEFAULT_MODEL_TYPE"),
        patch("google.adk.tools.FunctionTool"),
    ):
        agent = AnyAgent.create(AgentFramework.GOOGLE, AgentConfig(model_id="gpt-4o"))
        agent.run("foo", user_id="1", session_id="2", run_config=run_config)
        mock_runner.return_value.run_async.assert_called_once_with(
            user_id="1",
            session_id="2",
            new_message=types.Content(role="user", parts=[types.Part(text="foo")]),
            run_config=run_config,
        )



import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent


def test_create_any_with_framework(agent_framework: AgentFramework) -> None:
    agent = AnyAgent.create(agent_framework, AgentConfig(model_id="gpt-4o"))
    assert agent


def test_create_any_with_valid_string(agent_framework: AgentFramework) -> None:
    agent = AnyAgent.create(agent_framework.name, AgentConfig(model_id="gpt-4o"))
    assert agent


def test_create_any_with_invalid_string() -> None:
    with pytest.raises(ValueError, match="Unsupported agent framework"):
        AnyAgent.create("non-existing", AgentConfig(model_id="gpt-4o"))



import json

import pytest

from any_agent.evaluation.evaluation_case import CheckpointCriteria, EvaluationCase
from any_agent.tracing.trace import AgentSpan, AgentTrace


@pytest.fixture
def evaluation_case() -> EvaluationCase:
    return EvaluationCase(
        ground_truth=[{"name": "Test Case 1", "value": 1.0, "points": 1.0}],
        checkpoints=[
            CheckpointCriteria.model_validate(
                {"criteria": "Check if value is 1.0", "points": 1}
            )
        ],
        llm_judge="gpt-4o-mini",
        final_output_criteria=[],
    )


@pytest.fixture
def agent_trace() -> AgentTrace:
    trace_path = "tests/unit/evaluation/sample_traces/OPENAI.json"
    with open(trace_path, encoding="utf-8") as f:
        spans = json.loads(f.read())
    spans = [AgentSpan.model_validate_json(span) for span in spans]
    return AgentTrace(spans=spans)



from pydantic import BaseModel, ConfigDict


class EvaluationResult(BaseModel):
    """Represents the result of evaluating a criterion."""

    model_config = ConfigDict(extra="forbid")
    passed: bool
    reason: str
    criteria: str
    points: int



import json
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Protocol, assert_never

from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    SpanExporter,
    SpanExportResult,
)
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel

from any_agent import AgentFramework, TracingConfig
from any_agent.logging import logger

from .processors import TracingProcessor
from .trace import AgentSpan, AgentTrace

if TYPE_CHECKING:
    from opentelemetry.sdk.trace import ReadableSpan


class AnyAgentExporter(SpanExporter):
    """Build an `AgentTrace` and export to the different outputs."""

    def __init__(  # noqa: D107
        self,
        agent_framework: AgentFramework,
        tracing_config: TracingConfig,
    ):
        self.agent_framework = agent_framework
        self.tracing_config = tracing_config
        self.trace: AgentTrace = AgentTrace()
        self.processor: TracingProcessor | None = TracingProcessor.create(
            agent_framework
        )
        self.console: Console | None = None

        if self.tracing_config.console:
            self.console = Console()

    def print_to_console(self, span_kind: str, interaction: Mapping[str, Any]) -> None:
        """Print the span to the console."""
        if not self.console:
            msg = "Console is not initialized"
            raise RuntimeError(msg)
        style = getattr(self.tracing_config, span_kind.lower(), None)
        if not style or interaction == {}:
            return

        self.console.rule(span_kind, style=style)

        for key, value in interaction.items():
            if key == "output":
                self.console.print(
                    Panel(
                        Markdown(str(value or "")),
                        title="Output",
                    ),
                )
            else:
                self.console.print(f"{key}: {value}")

        self.console.rule(style=style)

    def export(self, spans: Sequence["ReadableSpan"]) -> SpanExportResult:  # noqa: D102
        if not self.processor:
            return SpanExportResult.SUCCESS

        for readable_span in spans:
            span = AgentSpan.from_readable_span(readable_span)
            try:
                span_kind, interaction = self.processor.extract_interaction(span)
                if span_kind == "LLM" and self.tracing_config.cost_info:
                    span.add_cost_info()

                self.trace.spans.append(span)

                if self.tracing_config.console and self.console:
                    self.print_to_console(span_kind, interaction)

            except (json.JSONDecodeError, TypeError, AttributeError) as e:
                logger.warning("Failed to parse span data, %s, %s", span, e)
                continue
        return SpanExportResult.SUCCESS


class Instrumenter(Protocol):  # noqa: D101
    def instrument(self, *, tracer_provider: TracerProvider) -> None: ...  # noqa: D102

    def uninstrument(self) -> None: ...  # noqa: D102


def get_instrumenter_by_framework(framework: AgentFramework) -> Instrumenter:
    """Get the instrumenter for the given agent framework."""
    if framework is AgentFramework.OPENAI:
        from openinference.instrumentation.openai_agents import (
            OpenAIAgentsInstrumentor,
        )

        return OpenAIAgentsInstrumentor()

    if framework is AgentFramework.SMOLAGENTS:
        from openinference.instrumentation.smolagents import SmolagentsInstrumentor

        return SmolagentsInstrumentor()

    if framework is AgentFramework.LANGCHAIN:
        from openinference.instrumentation.langchain import LangChainInstrumentor

        return LangChainInstrumentor()

    if framework is AgentFramework.LLAMA_INDEX:
        from openinference.instrumentation.llama_index import LlamaIndexInstrumentor

        return LlamaIndexInstrumentor()

    if (
        framework is AgentFramework.GOOGLE
        or framework is AgentFramework.AGNO
        or framework is AgentFramework.TINYAGENT
    ):
        msg = f"{framework} tracing is not supported."
        raise NotImplementedError(msg)

    assert_never(framework)



from textwrap import dedent

from any_agent.config import AgentFramework
from any_agent.evaluation.evaluation_case import EvaluationCase
from any_agent.evaluation.evaluators import (
    CheckpointEvaluator,
    HypothesisEvaluator,
    QuestionAnsweringSquadEvaluator,
)
from any_agent.evaluation.evaluators.schemas import EvaluationResult
from any_agent.evaluation.results_saver import save_evaluation_results
from any_agent.logging import logger
from any_agent.tracing import TracingProcessor
from any_agent.tracing.trace import AgentTrace


class EvaluationRunner:
    def __init__(self, output_path: str = "output") -> None:
        self._evaluation_cases: list[EvaluationCase] = []
        self._traces: list[AgentTrace] = []
        self._trace_frameworks: list[AgentFramework] = []
        self.checkpoint_evaluator: CheckpointEvaluator | None = None
        self.hypothesis_evaluator: HypothesisEvaluator | None = None
        self.qa_evaluator: QuestionAnsweringSquadEvaluator | None = None
        self.output_path: str = output_path

    def _setup_evaluators(self, evaluation_case: EvaluationCase) -> None:
        self.checkpoint_evaluator = CheckpointEvaluator(model=evaluation_case.llm_judge)
        self.hypothesis_evaluator = HypothesisEvaluator(model=evaluation_case.llm_judge)
        self.qa_evaluator = QuestionAnsweringSquadEvaluator()

    def add_evaluation_case(self, evaluation_case: EvaluationCase) -> None:
        """Add test case file path to the evaluation runner."""
        if evaluation_case not in self._evaluation_cases:
            self._evaluation_cases.append(evaluation_case)
        else:
            logger.warning("Test case %s already added.", evaluation_case)

    def add_trace(self, trace: AgentTrace, agent_framework: AgentFramework) -> None:
        """Add trace file path to the evaluation runner."""
        if trace not in self._traces:
            self._traces.append(trace)
            self._trace_frameworks.append(agent_framework)
        else:
            logger.warning("trace %salready added.", trace)

    def _run_trace_eval(
        self,
        evaluation_case: EvaluationCase,
        trace: AgentTrace,
        agent_framework: AgentFramework,
    ) -> None:
        processor = TracingProcessor.create(agent_framework)
        if not processor:
            msg = f"Processor for {agent_framework} not available."
            raise ValueError(msg)
        hypothesis_answer = processor._extract_hypothesis_answer(trace=trace)
        if not self.checkpoint_evaluator:
            msg = "CheckpointEvaluator not initialized."
            raise ValueError(msg)
        checkpoint_results = self.checkpoint_evaluator.evaluate(
            trace=trace,
            checkpoints=evaluation_case.checkpoints,
            processor=processor,
        )
        if not self.hypothesis_evaluator:
            msg = "HypothesisEvaluator not initialized."
            raise ValueError(msg)
        hypothesis_answer_results = self.hypothesis_evaluator.evaluate(
            hypothesis_final_output=hypothesis_answer,
            ground_truth_answer_dict=evaluation_case.ground_truth,
            ground_truth_checkpoints=evaluation_case.final_output_criteria,
        )

        if evaluation_case.ground_truth:
            if not self.qa_evaluator:
                msg = "QuestionAnsweringSquadEvaluator not initialized."
                raise ValueError(msg)
            direct_results = self.qa_evaluator.evaluate(
                hypothesis_answer=hypothesis_answer,
                ground_truth_answer=evaluation_case.ground_truth,
            )
        else:
            direct_results = []
        self._compile_results(
            evaluation_case=evaluation_case,
            trace=trace,
            hypothesis_answer=hypothesis_answer,
            checkpoint_results=checkpoint_results,
            hypothesis_answer_results=hypothesis_answer_results,
            direct_results=direct_results,
        )

    def _compile_results(
        self,
        evaluation_case: EvaluationCase,
        trace: AgentTrace,
        hypothesis_answer: str,
        checkpoint_results: list[EvaluationResult],
        hypothesis_answer_results: list[EvaluationResult],
        direct_results: list[EvaluationResult],
    ) -> None:
        verification_results = (
            checkpoint_results + hypothesis_answer_results + direct_results
        )

        output_message = ""
        output_message += f"""<yellow>Hypothesis Final answer extracted: {hypothesis_answer}</yellow>\n"""
        failed_checks = [r for r in verification_results if not r.passed]
        passed_checks = [r for r in verification_results if r.passed]
        missed_points = sum([r.points for r in failed_checks])
        won_points = sum([r.points for r in passed_checks])
        if passed_checks:
            for check in passed_checks:
                message = dedent(
                    f"""
                    <green>Passed:
                    - {check.criteria}
                    - {check.reason}</green>""",
                )
                output_message += message + "\n"
        if failed_checks:
            for check in failed_checks:
                message = dedent(
                    f"""
                    <red>Failed:
                    - {check.criteria}
                    - {check.reason}</red>""",
                )
                output_message += message + "\n"
        else:
            output_message += "<green>All checkpoints passed!</green>\n"
        output_message += f"<green>Passed checkpoints: {len(passed_checks)}</green>\n"
        output_message += f"<red>Failed checkpoints: {len(failed_checks)}</red>\n"
        output_message += "<green>=====================================</green>\n"
        output_message += (
            f"<green>Score: {won_points}/{won_points + missed_points}</green>\n"
        )
        output_message += "<green>=====================================</green>\n"
        logger.info(output_message)

        if won_points + missed_points == 0:
            msg = "No points were defined in the test case"
            raise ValueError(msg)
        score = won_points / (won_points + missed_points) * 100

        # Save the evaluation results
        save_evaluation_results(
            evaluation_case=evaluation_case,
            output_path=self.output_path,
            output_message=output_message,
            trace=trace,
            hypothesis_answer=hypothesis_answer,
            passed_checks=len(passed_checks),
            failed_checks=len(failed_checks),
            score=score,
        )

    def _run_evaluation_case(self, evaluation_case: EvaluationCase) -> None:
        self._setup_evaluators(evaluation_case)
        for trace, agent_framework in zip(
            self._traces, self._trace_frameworks, strict=True
        ):
            self._run_trace_eval(evaluation_case, trace, agent_framework)

    def run(self) -> None:
        """Run the evaluation for all test cases."""
        for evaluation_case in self._evaluation_cases:
            self._run_evaluation_case(evaluation_case)



from .config import AgentConfig, AgentFramework, TracingConfig
from .frameworks.any_agent import AnyAgent
from .tracing.trace import AgentTrace

__all__ = [
    "AgentConfig",
    "AgentFramework",
    "AgentTrace",
    "AnyAgent",
    "TracingConfig",
]



from .CheckpointEvaluator import CheckpointEvaluator
from .HypothesisEvaluator import HypothesisEvaluator
from .QuestionAnsweringSquadEvaluator import QuestionAnsweringSquadEvaluator

__all__ = [
    "CheckpointEvaluator",
    "HypothesisEvaluator",
    "QuestionAnsweringSquadEvaluator",
]



import os

import pytest
from litellm.utils import validate_environment

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.config import TracingConfig
from any_agent.tools import search_web, visit_webpage
from any_agent.tracing.trace import AgentTrace, _is_tracing_supported


@pytest.mark.skipif(
    os.environ.get("ANY_AGENT_INTEGRATION_TESTS", "FALSE").upper() != "TRUE",
    reason="Integration tests require `ANY_AGENT_INTEGRATION_TESTS=TRUE` env var",
)
def test_load_and_run_multi_agent(agent_framework: AgentFramework) -> None:
    kwargs = {}

    if agent_framework is AgentFramework.TINYAGENT:
        pytest.skip(
            f"Skipping test for {agent_framework.name} because it does not support multi-agent"
        )

    kwargs["model_id"] = "gpt-4.1-nano"
    env_check = validate_environment(kwargs["model_id"])
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    model_args = (
        {"parallel_tool_calls": False}
        if agent_framework is not AgentFramework.AGNO
        else None
    )
    main_agent = AgentConfig(
        instructions="You must use the available agents to complete the task.",
        description="The orchestrator that can use other agents.",
        model_args=model_args,
        **kwargs,  # type: ignore[arg-type]
    )

    managed_agents = [
        AgentConfig(
            name="search_web_agent",
            model_id="gpt-4.1-nano",
            description="Agent that can search the web",
            tools=[search_web],
            model_args=model_args,
        ),
        AgentConfig(
            name="visit_webpage_agent",
            model_id="gpt-4.1-nano",
            description="Agent that can visit webpages",
            tools=[visit_webpage],
            model_args=model_args,
        ),
    ]

    agent = AnyAgent.create(
        agent_framework=agent_framework,
        agent_config=main_agent,
        managed_agents=managed_agents,
        tracing=TracingConfig(console=False, cost_info=True),
    )
    agent_trace = agent.run("Which agent framework is the best?")

    assert agent_trace
    assert agent_trace.final_output
    if _is_tracing_supported(agent_framework):
        assert agent_trace.spans
        assert len(agent_trace.spans) > 0
        cost_sum = agent_trace.get_total_cost()
        assert cost_sum.total_cost > 0
        assert cost_sum.total_cost < 1.00
        assert cost_sum.total_tokens > 0
        assert cost_sum.total_tokens < 20000

    try:
        agent_trace = agent.run("Which agent framework is the best?")

        assert isinstance(agent_trace, AgentTrace)
        assert agent_trace.final_output
        if _is_tracing_supported(agent_framework):
            assert agent_trace.spans
            assert len(agent_trace.spans) > 0
            cost_sum = agent_trace.get_total_cost()
            assert cost_sum.total_cost > 0
            assert cost_sum.total_cost < 1.00
            assert cost_sum.total_tokens > 0
            assert cost_sum.total_tokens < 20000
    finally:
        agent.exit()



from abc import ABC, abstractmethod
from collections.abc import Sequence
from contextlib import AsyncExitStack
from textwrap import dedent
from typing import TYPE_CHECKING, Generic, Protocol, TypeVar, runtime_checkable

from pydantic import BaseModel, ConfigDict, PrivateAttr

from any_agent.config import MCPParams

if TYPE_CHECKING:
    from agents.mcp.server import MCPServer


@runtime_checkable
class HasName(Protocol):
    """Protocol for objects that have a name."""

    name: str


T = TypeVar("T")


class _MCPConnection(BaseModel, ABC, Generic[T]):
    mcp_tool: MCPParams
    _exit_stack: AsyncExitStack = PrivateAttr(default_factory=AsyncExitStack)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    async def list_tools(self) -> list[T]:
        """List tools from the MCP server."""

    @property
    def server(self) -> "MCPServer | None":
        """Return the MCP server instance."""
        return None

    def _filter_tools(self, tools: Sequence[T]) -> Sequence[T]:
        # Only add the tools listed in mcp_tool['tools'] if specified
        requested_tools = list(self.mcp_tool.tools or [])

        if not requested_tools:
            return tools

        tool_names = [
            tool.name if isinstance(tool, HasName) else tool for tool in tools
        ]

        found_tools = [tool for tool in tool_names if tool in requested_tools]

        if len(found_tools) != len(requested_tools):
            error_message = (
                dedent(
                    f"""Could not find all requested tools in the MCP server:
                    Requested ({len(requested_tools)}): {requested_tools}
                    Set ({len(tool_names)}):   {tool_names}
                """
                ),
            )
            raise ValueError(error_message)
        return tools



from collections.abc import Sequence
from typing import TYPE_CHECKING
from uuid import uuid4

from any_agent.config import AgentConfig, AgentFramework, TracingConfig
from any_agent.logging import logger
from any_agent.tools import search_web, visit_webpage

from .any_agent import AnyAgent

try:
    from google.adk.agents.llm_agent import LlmAgent
    from google.adk.models.lite_llm import LiteLlm
    from google.adk.runners import InMemoryRunner
    from google.adk.tools.agent_tool import AgentTool
    from google.genai import types

    DEFAULT_MODEL_TYPE = LiteLlm
    adk_available = True
except ImportError:
    adk_available = False

if TYPE_CHECKING:
    from google.adk.models.base_llm import BaseLlm

    from any_agent.tracing.trace import AgentTrace


class GoogleAgent(AnyAgent):
    """Google ADK agent implementation that handles both loading and running."""

    def __init__(
        self,
        config: AgentConfig,
        managed_agents: Sequence[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ):
        super().__init__(config, managed_agents, tracing)
        self._agent: LlmAgent | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.GOOGLE

    def _get_model(self, agent_config: AgentConfig) -> "BaseLlm":
        """Get the model configuration for a Google agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE
        return model_type(
            model=agent_config.model_id,
            api_key=agent_config.api_key,
            api_base=agent_config.api_base,
            **agent_config.model_args or {},
        )

    async def _load_agent(self) -> None:
        """Load the Google agent with the given configuration."""
        if not adk_available:
            msg = "You need to `pip install 'any-agent[google]'` to use this agent"
            raise ImportError(msg)

        if not self.managed_agents and not self.config.tools:
            self.config.tools = [
                search_web,
                visit_webpage,
            ]
        tools, _ = await self._load_tools(self.config.tools)

        sub_agents_instanced = []
        if self.managed_agents:
            for managed_agent in self.managed_agents:
                managed_tools, _ = await self._load_tools(managed_agent.tools)

                agent_type = managed_agent.agent_type or LlmAgent

                managed_agent_args = managed_agent.agent_args or {}
                handoff = managed_agent_args.pop("handoff", None)
                instance = agent_type(
                    name=managed_agent.name,
                    instruction=managed_agent.instructions or "",
                    model=self._get_model(managed_agent),
                    tools=managed_tools,
                    **managed_agent_args or {},
                )

                if handoff:
                    sub_agents_instanced.append(instance)
                else:
                    tools.append(AgentTool(instance))
        agent_type = self.config.agent_type or LlmAgent
        self._agent = agent_type(
            name=self.config.name,
            instruction=self.config.instructions or "",
            model=self._get_model(self.config),
            tools=tools,
            sub_agents=sub_agents_instanced,
            **self.config.agent_args or {},
            output_key="response",
        )

    async def run_async(  # type: ignore[no-untyped-def]
        self,
        prompt: str,
        user_id: str | None = None,
        session_id: str | None = None,
        **kwargs,
    ) -> "AgentTrace":
        """Run the Google agent with the given prompt."""
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        self._setup_tracing()
        runner = InMemoryRunner(self._agent)
        user_id = user_id or str(uuid4())
        session_id = session_id or str(uuid4())
        runner.session_service.create_session(
            app_name=runner.app_name,
            user_id=user_id,
            session_id=session_id,
        )
        events = runner.run_async(
            user_id=user_id,
            session_id=session_id,
            new_message=types.Content(role="user", parts=[types.Part(text=prompt)]),
            **kwargs,
        )

        async for event in events:
            logger.debug(event)
            if event.is_final_response():
                break

        session = runner.session_service.get_session(
            app_name=runner.app_name,
            user_id=user_id,
            session_id=session_id,
        )
        assert session, "Session should not be None"
        response = session.state.get("response", None)

        self._exporter.trace.final_output = response
        return self._exporter.trace



from __future__ import annotations

import yaml
from litellm.utils import validate_environment
from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import TypedDict


class CheckpointCriteria(BaseModel):
    """Represents a checkpoint criteria with a description."""

    model_config = ConfigDict(extra="forbid")
    criteria: str
    points: int


class GroundTruthAnswer(TypedDict):
    name: str
    value: float
    points: float


class EvaluationCase(BaseModel):
    model_config = ConfigDict(extra="forbid")
    ground_truth: list[GroundTruthAnswer] = Field(
        default_factory=list[GroundTruthAnswer],
    )
    checkpoints: list[CheckpointCriteria] = Field(
        default_factory=list[CheckpointCriteria],
    )
    llm_judge: str
    final_output_criteria: list[CheckpointCriteria] = Field(
        default_factory=list[CheckpointCriteria],
    )
    evaluation_case_path: str | None = None

    @classmethod
    def from_yaml(cls, evaluation_case_path: str) -> EvaluationCase:
        """Load a test case from a YAML file and process it."""
        with open(evaluation_case_path, encoding="utf-8") as f:
            evaluation_case_dict = yaml.safe_load(f)
        final_output_criteria = []

        def add_gt_final_output_criteria(
            ground_truth_list: list[GroundTruthAnswer],
        ) -> None:
            """Add checkpoints for each item in the ground_truth list."""
            for item in ground_truth_list:
                if "name" in item and "value" in item:
                    points = item.get(
                        "points",
                        1,
                    )  # Default to 1 if points not specified
                    final_output_criteria.append(
                        {
                            "points": points,
                            "criteria": f"Check if {item['name']} is approximately '{item['value']}'.",
                        },
                    )

        if "ground_truth" in evaluation_case_dict:
            add_gt_final_output_criteria(evaluation_case_dict["ground_truth"])
            evaluation_case_dict["final_output_criteria"] = final_output_criteria
            # remove the points from the ground_truth list but keep the name and value
            evaluation_case_dict["ground_truth"] = [
                item
                for item in evaluation_case_dict["ground_truth"]
                if isinstance(item, dict)
            ]

        evaluation_case_dict["evaluation_case_path"] = evaluation_case_path
        # verify that the llm_judge is a valid litellm model
        validate_environment(evaluation_case_dict["llm_judge"])
        return cls.model_validate(evaluation_case_dict)



from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.tools import (
    search_web,
    visit_webpage,
)


def test_load_llama_index_agent_default() -> None:
    model_mock = MagicMock()
    create_mock = MagicMock()
    agent_mock = MagicMock()
    create_mock.return_value = agent_mock
    tool_mock = MagicMock()
    from llama_index.core.tools import FunctionTool

    with (
        patch("any_agent.frameworks.llama_index.DEFAULT_AGENT_TYPE", create_mock),
        patch("any_agent.frameworks.llama_index.DEFAULT_MODEL_TYPE", model_mock),
        patch.object(FunctionTool, "from_defaults", tool_mock),
    ):
        AnyAgent.create(
            AgentFramework.LLAMA_INDEX,
            AgentConfig(
                model_id="gemini/gemini-2.0-flash",
                instructions="You are a helpful assistant",
            ),
        )

        model_mock.assert_called_once_with(
            model="gemini/gemini-2.0-flash", api_key=None, api_base=None
        )
        create_mock.assert_called_once_with(
            name="any_agent",
            llm=model_mock.return_value,
            system_prompt="You are a helpful assistant",
            description="The main agent",
            tools=[tool_mock(search_web), tool_mock(visit_webpage)],
        )


def test_load_llama_index_agent_missing() -> None:
    with patch("any_agent.frameworks.llama_index.llama_index_available", False):
        with pytest.raises(ImportError):
            AnyAgent.create(AgentFramework.LLAMA_INDEX, AgentConfig(model_id="gpt-4o"))


def test_load_llama_index_multiagent() -> None:
    model_mock = MagicMock()
    create_mock = MagicMock()
    agent_mock = MagicMock()
    create_mock.return_value = agent_mock
    tool_mock = MagicMock()
    from llama_index.core.tools import FunctionTool

    with (
        patch("any_agent.frameworks.llama_index.DEFAULT_AGENT_TYPE", create_mock),
        patch("any_agent.frameworks.llama_index.AgentWorkflow"),
        patch("any_agent.frameworks.llama_index.DEFAULT_MODEL_TYPE", model_mock),
        patch.object(FunctionTool, "from_defaults", tool_mock),
    ):
        main_agent = AgentConfig(model_id="gpt-4.1-mini")

        managed_agents = [
            AgentConfig(
                model_id="gpt-4.1-nano",
                tools=[
                    search_web,
                    visit_webpage,
                ],
            ),
        ]

        AnyAgent.create(
            AgentFramework.LLAMA_INDEX, main_agent, managed_agents=managed_agents
        )

        create_mock.assert_any_call(
            name="managed_agent_0",
            description="A managed agent",
            llm=model_mock.return_value,
            tools=[
                tool_mock(search_web),
                tool_mock(visit_webpage),
            ],
            system_prompt=None,
            can_handoff_to=["any_agent"],
        )

        create_mock.assert_called_with(
            name="any_agent",
            description="The main agent",
            llm=model_mock.return_value,
            system_prompt=None,
            can_handoff_to=["managed_agent_0"],
            tools=[],
        )


def test_run_llama_index_agent_custom_args() -> None:
    create_mock = MagicMock()
    agent_mock = AsyncMock()
    create_mock.return_value = agent_mock
    from llama_index.core.tools import FunctionTool

    with (
        patch("any_agent.frameworks.llama_index.DEFAULT_AGENT_TYPE", create_mock),
        patch("any_agent.frameworks.llama_index.DEFAULT_MODEL_TYPE"),
        patch.object(FunctionTool, "from_defaults"),
    ):
        agent = AnyAgent.create(
            AgentFramework.LLAMA_INDEX,
            AgentConfig(
                model_id="gemini/gemini-2.0-flash",
                instructions="You are a helpful assistant",
            ),
        )
        agent.run("foo", timeout=10)
        agent_mock.run.assert_called_once_with("foo", timeout=10)



from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, cast

from any_agent import AgentConfig, AgentFramework
from any_agent.config import TracingConfig
from any_agent.logging import logger
from any_agent.tools import search_web, visit_webpage

from .any_agent import AnyAgent

try:
    from llama_index.core.agent.workflow import AgentWorkflow, ReActAgent
    from llama_index.llms.litellm import LiteLLM

    DEFAULT_AGENT_TYPE = ReActAgent
    DEFAULT_MODEL_TYPE = LiteLLM
    llama_index_available = True
except ImportError:
    llama_index_available = False


if TYPE_CHECKING:
    from llama_index.core.agent.workflow.workflow_events import AgentOutput
    from llama_index.core.llms import LLM

    from any_agent.tracing.trace import AgentTrace


class LlamaIndexAgent(AnyAgent):
    """LLamaIndex agent implementation that handles both loading and running."""

    def __init__(
        self,
        config: AgentConfig,
        managed_agents: Sequence[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ):
        super().__init__(config, managed_agents, tracing)
        self._agent: AgentWorkflow | ReActAgent | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.LLAMA_INDEX

    def _get_model(self, agent_config: AgentConfig) -> "LLM":
        """Get the model configuration for a llama_index agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE
        return cast(
            "LLM",
            model_type(
                model=agent_config.model_id,
                api_key=agent_config.api_key,
                api_base=agent_config.api_base,
                **agent_config.model_args or {},
            ),
        )

    async def _load_agent(self) -> None:
        """Load the LLamaIndex agent with the given configuration."""
        if not llama_index_available:
            msg = "You need to `pip install 'any-agent[llama_index]'` to use this agent"
            raise ImportError(msg)

        if not self.managed_agents and not self.config.tools:
            self.config.tools = [
                search_web,
                visit_webpage,
            ]

        if self.managed_agents:
            agents = []
            managed_names = []
            for n, managed_agent in enumerate(self.managed_agents):
                managed_tools, _ = await self._load_tools(managed_agent.tools)
                name = managed_agent.name
                if not name or name == "any_agent":
                    logger.warning(
                        "Overriding name for managed_agent. Can't use the default.",
                    )
                    name = f"managed_agent_{n}"
                managed_names.append(name)
                agent_type = managed_agent.agent_type or DEFAULT_AGENT_TYPE
                managed_instance = agent_type(
                    name=name,
                    description=managed_agent.description or "A managed agent",
                    system_prompt=managed_agent.instructions,
                    tools=managed_tools,
                    llm=self._get_model(managed_agent),
                    can_handoff_to=[self.config.name],
                    **managed_agent.agent_args or {},
                )
                agents.append(managed_instance)

            main_tools, _ = await self._load_tools(self.config.tools)
            agent_type = self.config.agent_type or DEFAULT_AGENT_TYPE
            main_agent = agent_type(
                name=self.config.name,
                description=self.config.description or "The main agent",
                tools=main_tools,
                llm=self._get_model(self.config),
                system_prompt=self.config.instructions,
                can_handoff_to=managed_names,
                **self.config.agent_args or {},
            )
            agents.append(main_agent)

            self._agent = AgentWorkflow(agents=agents, root_agent=main_agent.name)

        else:
            imported_tools, _ = await self._load_tools(self.config.tools)
            agent_type = self.config.agent_type or DEFAULT_AGENT_TYPE
            self._agent = agent_type(
                name=self.config.name,
                tools=imported_tools,
                description=self.config.description or "The main agent",
                llm=self._get_model(self.config),
                system_prompt=self.config.instructions,
                **self.config.agent_args or {},
            )

    async def run_async(self, prompt: str, **kwargs: Any) -> "AgentTrace":
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        self._setup_tracing()
        result: AgentOutput = await self._agent.run(prompt, **kwargs)
        # assert that it's a TextBlock
        if not result.response.blocks or not hasattr(result.response.blocks[0], "text"):
            msg = f"Agent did not return a valid response: {result.response}"
            raise ValueError(msg)

        self._exporter.trace.final_output = result.response.blocks[0].text
        return self._exporter.trace



import os

import pandas as pd

from any_agent.evaluation.evaluation_case import EvaluationCase
from any_agent.logging import logger
from any_agent.tracing.trace import AgentTrace

# Use the shared logger


def save_evaluation_results(
    evaluation_case: EvaluationCase,
    output_path: str,
    output_message: str,
    trace: AgentTrace,
    hypothesis_answer: str,
    passed_checks: int,
    failed_checks: int,
    score: float,
) -> None:
    """Save evaluation results to the specified output path.

    Args:
        evaluation_case: Path to the test case file
        agent_config: Path to the agent configuration file
        output_path: Path to save the results
        output_message: Formatted output message with evaluation details
        trace: The trace
        hypothesis_answer: The extracted hypothesis answer
        passed_checks: Number of passed checkpoints
        failed_checks: Number of failed checkpoints
        score: Evaluation score as a percentage

    """
    # Ensure the output directory exists
    os.makedirs(os.path.dirname(output_path), exist_ok=True)

    # See if the output_path file exists
    if os.path.exists(output_path):
        logger.info(f"Reading existing output from {output_path}")
        df = pd.read_json(output_path, orient="records", lines=True)
    else:
        logger.info(f"Creating new output file at {output_path}")
        df = pd.DataFrame()

    df = pd.concat(
        [
            df,
            pd.DataFrame(
                [
                    {
                        "config": evaluation_case.model_dump(),
                        "evaluation_case_path": evaluation_case.evaluation_case_path,
                        "output_message": output_message,
                        "trace": trace.model_dump_json(),
                        "hypothesis_answer": hypothesis_answer,
                        "passed_checks": passed_checks,
                        "failed_checks": failed_checks,
                        "score": round(score, 2),
                    },
                ],
            ),
        ],
    )
    logger.info(f"Writing output to {output_path}")
    df.to_json(output_path, orient="records", lines=True)



import shutil
from collections.abc import AsyncGenerator, Generator, Sequence
from contextlib import AsyncExitStack
from typing import Any, Protocol
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from mcp import Tool as MCPTool
from mcp.client.session import ClientSession
from pydantic import Field
from pytest_lazy_fixtures import lf

from any_agent.config import MCPParams, MCPSse, MCPStdio, Tool
from any_agent.tools import _MCPConnection


class Toolset(Protocol):
    def load_tools(self) -> list[Tool]: ...


@pytest.fixture
def tools() -> list[Tool]:
    return ["write_file", "read_file", "other_tool"]


@pytest.fixture
def mcp_sse_params_no_tools() -> MCPSse:
    return MCPSse(
        url="http://localhost:8000/sse",
        headers={"Authorization": "Bearer test-token"},
    )


@pytest.fixture
def session() -> Generator[Any]:
    return AsyncMock()


@pytest.fixture
def _path_client_session(session: AsyncGenerator[Any]) -> Generator[None]:
    with patch(
        "any_agent.tools.mcp.frameworks.agno.ClientSession"
    ) as mock_client_session:
        mock_client_session.return_value.__aenter__.return_value = session
        yield


@pytest.fixture
def mcp_sse_params_with_tools(
    mcp_sse_params_no_tools: MCPSse, tools: Sequence[Tool]
) -> MCPSse:
    return mcp_sse_params_no_tools.model_copy(update={"tools": tools})


@pytest.fixture
def enter_context_with_transport_and_session(
    session: Any,
    tools: Sequence[str],
) -> Generator[None]:
    transport = (AsyncMock(), AsyncMock())
    with patch.object(AsyncExitStack, "enter_async_context") as mock_context:
        mock_context.side_effect = [transport, session, tools]
        yield


@pytest.fixture
def command() -> str:
    # Mocking the command part of stdio is really tricky so instead we'll use
    # a real command that should be available on all systems (this is what openai-agents does too)
    tee = shutil.which("tee") or ""
    assert tee, "tee not found"
    return tee


@pytest.fixture
def stdio_params(command: str, tools: Sequence[str]) -> MCPStdio:
    return MCPStdio(command=command, args=[], tools=tools)


@pytest.fixture
def mcp_tools(tools: Sequence[str]) -> list[MCPTool]:
    return [
        MCPTool(name=tool, inputSchema={"type": "string", "properties": {}})
        for tool in tools
    ]


@pytest.fixture
def _patch_client_session_initialize() -> Generator[ClientSession]:
    with patch(
        "mcp.client.session.ClientSession.initialize",
        new_callable=AsyncMock,
        return_value=None,
    ) as initialize_mock:
        yield initialize_mock


@pytest.fixture
def _patch_client_session_list_tools(mcp_tools: Sequence[MCPTool]) -> Generator[None]:
    tool_list = MagicMock()
    tool_list.tools = mcp_tools
    with patch("mcp.client.session.ClientSession.list_tools", return_value=tool_list):
        yield


@pytest.fixture
def sse_params_echo_server(echo_sse_server: Any, tools: Sequence[str]) -> MCPSse:
    return MCPSse(url=echo_sse_server["url"], tools=tools)


class FakeMCPConnection(_MCPConnection[Any]):
    mcp_tool: None = None  # type: ignore[assignment]
    tools: Sequence[Tool] = Field(default_factory=list)

    async def list_tools(self) -> list[Tool]:
        return list(self.tools)


@pytest.fixture
def mcp_connection(tools: Sequence[Tool]) -> _MCPConnection[Any]:
    return FakeMCPConnection(tools=tools)


@pytest.fixture(
    params=[lf("stdio_params"), lf("mcp_sse_params_with_tools")], ids=["STDIO", "SSE"]
)
def mcp_params(request: pytest.FixtureRequest) -> MCPParams:
    return request.param  # type: ignore[no-any-return]
